/*
 * Morfeo Project
 * http://morfeo-project.org/
 * 
 * Copyright (C) 2008 Telefonica Investigacion y Desarrollo S.A. Unipersonal
 * 
 * Author      : Alvaro Polo Valdenebro <apv at tid.es>
 * Module      : OverloadedServerImpl
 * Description : Implementation for overloaded server servant object
 *
 * Info about members and contributors of the MORFEO project
 * is available at:
 *
 *     http://www.morfeo-project.org/TIDorbJ/CREDITS
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * If you want to use this software an plan to distribute a
 * proprietary application in any way, and you are not licensing and
 * distributing your source code under GPL, you probably need to
 * purchase a commercial license of the product.  More info about
 * licensing options is available at:
 *
 *   http://www.morfeo-project.org/TIDorbJ/Licensing
 *
 */

import Overload.*;
import Overload.OverloadedServerPackage.Statistics;

public class OverloadedServerImpl extends OverloadedServerPOA
{

   private long m_start_timer;
   private long m_stop_timer;
   
   private long m_expected_requests;
   private long m_processed_requests;
   private long m_accounted_requests;
   private long m_ignored_requests;
   private double m_flood_threshold;
  
   protected static void compute(StructType value)
   {
      // Do some computing
      // value.string_val = new String("Foobar string");
      for (int i = 0; i < 50000; i++)
      {
         value.long_val += (int)(i * 14234.2532);
         value.long_val += (int)(i / 45.354 + 10);
      }
   }
   
   protected static void compute(ComplexStructType value)
   {
      // Do some computing
      // value.string_val = new String("Foobar string");
      for (int i = 0; i < 500; i++)
      {
         value.long_val = (int)(i * 14234.2532);
         value.long_val = (int)(i / 45.354 + 10);
      }
   }
   
   protected synchronized void request()
   {
      m_processed_requests++;
      long lower_thres = (long) (m_expected_requests * m_flood_threshold);
      long upper_thres = m_expected_requests - lower_thres;
      if (m_ignored_requests < lower_thres)
      {
         // In lower flood threshold
         m_ignored_requests++;
      }
      else if (m_ignored_requests + m_accounted_requests >= upper_thres)
      {
         // In upper flood threshold
         m_ignored_requests++;
         if (m_stop_timer == 0)
            m_stop_timer = System.currentTimeMillis();
      }
      else
      {
         if (m_start_timer == 0)
            m_start_timer = System.currentTimeMillis();
         
         // When flood threshold is zero, the last request should
         // set the stop timer
         if (m_flood_threshold == 0.0 && 
             m_processed_requests == m_expected_requests)
            m_stop_timer = System.currentTimeMillis();
         
         m_accounted_requests++;
      }
   }
   
   protected synchronized long getExpectedRequests()
   {
      return m_expected_requests;
   }
   
   protected synchronized long getProcessedRequests()
   {
      return m_processed_requests;
   }
   
   protected synchronized long getAccountedRequests()
   {
      return m_accounted_requests;
   }
   
   protected synchronized long getIgnoredRequests()
   {
      return m_ignored_requests;
   }
   
   protected synchronized long getMilliseconds()
   {
      return m_stop_timer - m_start_timer;
   }
   
   protected synchronized float getRequestsPerSecond()
   {
      return (float) (getAccountedRequests() / (getMilliseconds() / 1000.0));
   }
   
   /* Constructor */
   public OverloadedServerImpl()
   {
      m_start_timer        = 0;
      m_stop_timer         = 0;
      
      m_expected_requests  = 0;
      m_processed_requests = 0;
      m_accounted_requests = 0;      
      m_ignored_requests   = 0;
      m_flood_threshold    = 0.0;
   }
   
   /***
    * Operations for measurement
    */
   public synchronized void reset(double flood_threshold, long requests)
   {
      m_start_timer = 0;
      m_stop_timer  = 0;
      
      m_expected_requests = requests;
      m_processed_requests = 0;      
      m_accounted_requests = 0;      
      m_ignored_requests = 0;      
      m_flood_threshold = flood_threshold;
      
      System.err.println(
            "Reset: " + requests + " reqs, " + 
            flood_threshold * 100.0 + "% ignored (first and last " + 
            (long) (requests * flood_threshold) + ")");
   }
   
   public synchronized Statistics getStats()
   {
      if (m_stop_timer == 0)
         m_stop_timer = System.currentTimeMillis();
      
      return new Statistics(
            getExpectedRequests(),
            getProcessedRequests(),
            getAccountedRequests(),
            getIgnoredRequests(),
            getMilliseconds(),
            getRequestsPerSecond());
   }
   
   /***
    * Operations for in, out, inout and return of not-too-complex values
    */
   public void doSomethingIn(StructType value)
   {
      request();
      compute(value);
   }

   public void doSomethingOut(StructTypeHolder value)
   {
      request();
      value.value = new StructType();
      compute(value.value);
   }

   public void doSomethingInOut(StructTypeHolder value)
   {
      request();
      compute(value.value);
   }

   public StructType doSomethingReturn()
   {
      request();
      StructType result = new StructType();
      compute(result);
      return result;
   }

   /***
    * Operations for in, out, inout and return of not-too-simple values
    */
   public void doSomethingComplexIn(ComplexStructType value)
   {
      request();
      compute(value);
   }

   public void doSomethingComplexOut(ComplexStructTypeHolder value)
   {
      request();
      value.value = new ComplexStructType();
      compute(value.value);
   }

   public void doSomethingComplexInOut(ComplexStructTypeHolder value)
   {
      request();
      compute(value.value);
   }

   public ComplexStructType doSomethingComplexReturn()
   {
      request();
      ComplexStructType result = new ComplexStructType();
      compute(result);
      return result;
   }

}
